-------------------------------------------------------------------------------
--! @file       AEAD_Core_FullWrapper.vhd
--! @brief      5-bit wrapper for AEAD_Core.vhd
--! @project    CAESAR Candidate Evaluation
--! @author     Ekawat (ice) Homsirikamol
--! @copyright  Copyright (c) 2015 Cryptographic Engineering Research Group
--!             ECE Department, George Mason University Fairfax, VA, U.S.A.
--!             All rights Reserved.
--! @license    This project is released under the GNU Public License.
--!             The license and distribution terms for this file may be
--!             found in the file LICENSE in this distribution or at
--!             http://www.gnu.org/licenses/gpl-3.0.txt
--! @note       This is publicly available encryption source code that falls
--!             under the License Exception TSU (Technology and software-
--!             —unrestricted)
-------------------------------------------------------------------------------

library ieee;
use ieee.std_logic_1164.all;

entity AEAD_Core_FullWrapper is
    generic (
        G_W             : integer := 256;
        G_SW            : integer := 32
    );
    port (
        --! Global signals
        clk             : in  std_logic;
        rst             : in  std_logic;

        --! SERDES signals
        sin             : in  std_logic;
        ssel            : in  std_logic;
        sout            : out std_logic
    );
end entity AEAD_Core_FullWrapper;

architecture structure of AEAD_Core_FullWrapper is
    signal sipo         : std_logic_vector(3*G_W+G_SW+8         -1 downto 0);
    signal piso         : std_logic_vector(2*G_W+9              -1 downto 0);
    signal piso_data    : std_logic_vector(2*G_W+9              -1 downto 0);
begin

    pReg:
    process(clk)
    begin
        if rising_edge(clk) then
            sipo <= sin & sipo(3*G_W+G_SW+8-1 downto 1);
            if (ssel = '1') then
                piso <= piso_data;
            else
                piso <= '0' & piso(2*G_W+9-1  downto 1);
            end if;
        end if;
    end process;
    sout <= piso(0);

    u_crypto:
    entity work.AEAD_Core(structure)
    generic map (
        G_W                     => G_W,
        G_SW                    => G_SW
    )
    port map (
        clk                     => clk      ,
        rst                     => rst      ,
                
        --! Input signals
        pdi                     => sipo(       G_W*1-1 downto           0),
        sdi                     => sipo(  G_SW+G_W*1-1 downto       G_W  ),
        bypass_fifo_data        => sipo(  G_SW+G_W*2-1 downto  G_SW+G_W  ),
        aux_fifo_dout           => sipo(  G_SW+G_W*3-1 downto  G_SW+G_W*2),
        pdi_valid               => sipo(1+G_SW+G_W*3-1),
        sdi_valid               => sipo(2+G_SW+G_W*3-1),
        do_ready                => sipo(3+G_SW+G_W*3-1),
        bypass_fifo_full        => sipo(4+G_SW+G_W*3-1),
        bypass_fifo_empty       => sipo(5+G_SW+G_W*3-1),
        aux_fifo_status         => sipo(8+G_SW+G_W*3-1 downto 5+G_SW+G_W*3),
        
        --! Output signals
        do                      => piso_data(  G_W  -1 downto     0  ),
        aux_fifo_din            => piso_data(  G_W*2-1 downto   G_W  ),
        pdi_ready               => piso_data(1+G_W*2-1),
        sdi_ready               => piso_data(2+G_W*2-1),
        do_valid                => piso_data(3+G_W*2-1),
        bypass_fifo_wr          => piso_data(4+G_W*2-1),
        bypass_fifo_rd          => piso_data(5+G_W*2-1),
        aux_fifo_ctrl           => piso_data(9+G_W*2-1 downto 5+G_W*2)
    );
end structure;