-------------------------------------------------------------------------------
--! @file       AEAD_Core_Wrapper.vhd
--! @brief      Wrapper for memory region of AEAD_Core
--! @project    CAESAR Candidate Evaluation
--! @author     Ekawat (ice) Homsirikamol
--! @copyright  Copyright (c) 2015 Cryptographic Engineering Research Group
--!             ECE Department, George Mason University Fairfax, VA, U.S.A.
--!             All rights Reserved.
--! @license    This project is released under the GNU Public License.
--!             The license and distribution terms for this file may be
--!             found in the file LICENSE in this distribution or at
--!             http://www.gnu.org/licenses/gpl-3.0.txt
--! @note       This is publicly available encryption source code that falls
--!             under the License Exception TSU (Technology and software-
--!             —unrestricted)
-------------------------------------------------------------------------------

library ieee;
use ieee.std_logic_1164.all;

entity AEAD_Core_Wrapper is
    generic (
        G_W             :   integer := 256;
        G_SW            :   integer := 32
    );
    port (
        --! Global signals
        clk             :   in  std_logic;
        rst             :   in  std_logic;

        --! Data in signals
        pdi             :   in  std_logic_vector(G_W            -1 downto 0);
        pdi_valid       :   in  std_logic;
        pdi_ready       :   out std_logic;

        --! Key signals
        sdi             :   in  std_logic_vector(G_SW           -1 downto 0);
        sdi_valid       :   in  std_logic;
        sdi_ready       :   out std_logic;

        --! Data out signals
        do              :   out std_logic_vector(G_W            -1 downto 0);
        do_ready        :   in  std_logic;
        do_valid        :   out std_logic;

        --! SERDES signals
        sin             :   in  std_logic;
        ssel            :   in  std_logic;
        sout            :   out std_logic
    );
end entity AEAD_Core_Wrapper;

architecture structure of AEAD_Core_Wrapper is
    signal sipo         : std_logic_vector(2*G_W+G_SW+5         -1 downto 0);
    signal piso         : std_logic_vector(  G_W+6              -1 downto 0);
    signal piso_data    : std_logic_vector(  G_W+6              -1 downto 0);
begin

    pReg:
    process(clk)
    begin
        if rising_edge(clk) then
            sipo <= sin & sipo(2*G_W+G_SW+5-1 downto 1);
            if (ssel = '1') then
                piso <= piso_data;
            else
                piso <= '0' & piso(G_W+6-1  downto 1);
            end if;
        end if;
    end process;
    sout <= piso(0);

    u_crypto:
    entity work.AEAD_Core(structure)
    generic map (
        G_W                     => G_W,
        G_SW                    => G_SW
    )
    port map (
        clk                     => clk      ,
        rst                     => rst      ,
        pdi                     => pdi      ,
        pdi_valid               => pdi_valid,
        pdi_ready               => pdi_ready,
        sdi                     => sdi      ,
        sdi_valid               => sdi_valid,
        sdi_ready               => sdi_ready,
        do                      => do       ,
        do_ready                => do_ready ,
        do_valid                => do_valid ,

        --! FIFO signals
        bypass_fifo_data        => sipo(  G_W  -1 downto       0),
        aux_fifo_dout           => sipo(  G_W*2-1 downto   G_W  ),
        bypass_fifo_full        => sipo(1+G_W*2-1),
        bypass_fifo_empty       => sipo(2+G_W*2-1),
        aux_fifo_status         => sipo(5+G_W*2-1 downto 2+G_W*2),
        
        aux_fifo_din            => piso_data(  G_W-1 downto     0),
        aux_fifo_ctrl           => piso_data(4+G_W-1 downto   G_W),
        bypass_fifo_wr          => piso_data(5+G_W-1),
        bypass_fifo_rd          => piso_data(6+G_W-1)
    );
end structure;